/* -*- mode: c++; -*- */
#ifndef _IMAGE_
#define _IMAGE_


#include <stdio.h>
#include <stdlib.h>

extern "C"
{
#ifdef __WIN32__
#include "jpeg/jpeglib.h"
#include "jpeg/jerror.h"
#else
#include <jpeglib.h>
#include <jerror.h>
#endif
}

#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string>
#include <cassert>
#include <cmath>
#include <cstring>

class Image
{

protected:

  int width ;
  int height ;
  int depth ;
  int components ;
  float mm_width ;
  float mm_height ;
  float mm_depth ;
  unsigned char * buffer ;
  //static 
public:

  Image ( int w, int h, int c )
  {
    this->width = w ;
    this->height = h ;
    this->components = c ;
    this->buffer = new unsigned char [ w * h * c ] ;
  }


  Image ()
  {
    this->width = this->height = 0 ;
    this->mm_width = this->mm_height = this->mm_depth = 0.0f ;
    this->buffer = NULL ;
  }


  virtual ~Image ()
  {
    delete [] buffer ;
  }

  int getWidth () { return this->width ; }
  int getHeight () { return this->height ; }

  unsigned char * getPixels () { return this->buffer ; }

  unsigned int getPixelArea ( float x1, float y1,
			      float x2, float y2,
			      unsigned int keyColour ) ;

  unsigned char getPixelRed ( int x, int y )
  {
    return (unsigned int) this->buffer [ ( y * width + x ) * 3 + 0 ] ;
  }

  unsigned char getPixelGreen( int x, int y )
  {
    return (unsigned int) this->buffer [ ( y * width + x ) * 3 + 1 ] ;
  }

  unsigned char getPixelBlue ( int x, int y )
  {
    return (unsigned int) this->buffer [ ( y * width + x ) * 3 + 2 ] ;
  }

  unsigned int getPixel ( float x, float y )
  {
    return getPixel ( (int) x, (int) y ) ;
  }

  unsigned int getPixel ( int x, int y )
  {
    int p = ( y * width + x ) * 3 ;

    return ( (unsigned int) this->buffer [ p + 0 ] << 24 ) +
      ( (unsigned int) this->buffer [ p + 1 ] << 16 ) +
      ( (unsigned int) this->buffer [ p + 2 ] << 8 ) + 255 ;
  }

  void setPixel ( float x, float y, unsigned int rgba )
  {
    setPixel ( (int) x, (int) y, rgba ) ;
  }

  void setPixel ( int x, int y, unsigned int rgba )
  {
    int p = ( y * this->width + x ) * 3 ;

    if ( ( rgba & 0xFF ) == 0 )
      return ;

    if ( ( rgba & 0xFF ) == 255 ) {
	this->buffer [ p + 0 ] = ( rgba >> 24 ) & 0xFF ;
	this->buffer [ p + 1 ] = ( rgba >> 16 ) & 0xFF ;
	this->buffer [ p + 2 ] = ( rgba >> 8 ) & 0xFF ;
      } else {
	unsigned int r = ( rgba >> 24 ) & 0xFF ;
	unsigned int g = ( rgba >> 16 ) & 0xFF ;
	unsigned int b = ( rgba >> 8 ) & 0xFF ;
	unsigned int a = ( rgba >> 0 ) & 0xFF ;
	unsigned int ac = 255 - a ;

	this->buffer [ p + 0 ] = (unsigned char)((int)(buffer [ p + 0 ]) * ac /
						 255 + r * a / 255 ) ;
	this->buffer [ p + 1 ] = (unsigned char)((int)(buffer [ p + 1 ]) * ac /
						 255 + g * a / 255 ) ;
	this->buffer [ p + 2 ] = (unsigned char)((int)(buffer [ p + 2 ]) * ac /
						 255 + b * a / 255 ) ;
      }
  }

  static void setHorizAvg ( int newh ) {} ;
  static void setVertAvg ( int newv ) {} ;

  void copyBuffer ( unsigned char * buf, int beg, int size ) {
    for ( int i = beg ; i < size ; ++ i )
      this->buffer[ i ] = buf[ i ] ;
  } ;

  virtual int load ( const char * filename ) = 0 ;
  virtual int save ( const char * filename ) = 0 ;

} ;


#endif
