#include <cstdlib>
#include <iostream>

#include "v3d.hpp"
#include "jpg.hpp"

using namespace std;


int
JPEG::load ( const char * filename )
{
  jpeg_decompress_struct cinfo ;
  jpeg_error_mgr jerr ;

  FILE *fd ;
  JSAMPARRAY linebuffer ;
  int row_stride ;

  if ( ( fd = fopen ( filename, "rb" ) ) == NULL )
    {
      perror ( "load JPEG" ) ;
      cerr << "load JPEG: Nie mog otworzy " << filename <<
	" do odczytu." << endl ;
      return 0 ;
    }


  cinfo.err = jpeg_std_error ( &jerr ) ;
  jpeg_create_decompress ( &cinfo ) ;
  jpeg_stdio_src ( &cinfo, fd ) ;
  jpeg_read_header ( &cinfo, TRUE ) ;
  jpeg_start_decompress ( &cinfo ) ;

  row_stride = cinfo.output_width * cinfo.output_components ;
  this->components = cinfo.output_components ;
  this->colorspace = cinfo.out_color_space ; //out_color_space

  /*
  if ( cinfo.output_components != 3 )
    {
      cerr << "readJPEG: " << filename << " has " <<
	cinfo.output_components << " components?!?" << endl ;
      return 0 ;
    }
  */

  linebuffer = ( * cinfo.mem->alloc_sarray )
    ( (j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1 ) ;

  delete [] buffer ;

  this->height = cinfo.output_height ;
  this->width = cinfo.output_width ;

  this->buffer = new unsigned char [ height * row_stride ] ;

  while (cinfo.output_scanline < cinfo.output_height)
    {
      int y = cinfo.output_scanline ;

      /*
	 jpeg_read_scanlines expects an array of pointers to scanlines.
	 Here the array is only one element long, but you could ask for
	 more than one scanline at a time if that's more convenient.
      */

      jpeg_read_scanlines ( &cinfo, linebuffer, 1 ) ;
      memcpy ( & buffer [ y * row_stride ], linebuffer[0], row_stride ) ;
    }

  jpeg_finish_decompress ( &cinfo ) ;
  jpeg_destroy_decompress ( &cinfo ) ;

  fclose ( fd ) ;
  return 1 ;
}

//processRawFrame ( const char * fname , const char * foutname  )
int
JPEG::RGBtoGrayscale( void )
{

  if ( this->colorspace == JCS_GRAYSCALE && this->components == 1 ) {
    cerr << "RGBtoGrayscale: Nic do roboty." << endl ;
    return 0 ;
  } else if ( this->colorspace != JCS_RGB ) {
    cerr << "RGBtoGrayscale: Nie-RGB (" << this->colorspace << ") !" << endl ;
    return 1 ;
    //  jeli ma by konwersja g16 -> jpeg
    //  } else if ( this->components == 2 ) {
  } else if ( this->components != 3 ) {
    cerr << "RGBtoGrayscale: Za warto bpp!" << endl ;
    return 2 ;
  }

  //  float min = 65536.0f , max = 0.0f ;
  unsigned char * buff_g = new unsigned char [ this->width * this->height ] ;

  cerr << " Szeroko: " << this->width << " Wysoko: " << this->height <<
    /*" max: " << max << " min: " << min <<*/ endl ;

  for ( int i = 0 ; i < this->width ; ++ i )
    for ( int j = 0 ; j < this->height ; ++ j )
      {
	// Get Red
	unsigned char pr = ( unsigned char ) this->buffer
	  [ ( j * this->width + i ) * 3 + 0 ] ;
	// Get Green
	unsigned char pg = ( unsigned char ) this->buffer
	  [ ( j * this->width + i ) * 3 + 1 ] ;
	// Get Blue
	unsigned char pb = ( unsigned char ) this->buffer
	  [ ( j * this->width + i ) * 3 + 2 ] ;

	if ( pr != pg || pg != pb ) {
	  unsigned int gray ; //,  px ;

	  /*
	    How do you convert color images to grayscale?
	    My first answer was to just average the red, green,
	    and blue values. That answer is wrong.
	    The reality is that the human eye is much better at seeing
	    some colors than others.
	    To get an accurate grayscale representation,
	    you need to apply different coefficients to the color samples.
	    Appropriate coefficients are 0.299 for red, 0.587 for green,
	    and 0.114 for blue.
	  */

	  gray = ( unsigned int ) ( pr * 0.299 + pg * 0.587 + pb * 0.114 ) ;

	  if ( gray > 255 ) {
	    cout << "!" << endl ;
	    gray = 255 ;
	  }

	  buff_g[ j * this->width + i ] = ( unsigned char ) gray ;

	} else {

	  buff_g[ j * this->width + i ] = ( unsigned char ) pr ;

	}
      }

  delete [] this->buffer ;
  this->buffer = buff_g ;
  this->components = 1 ;
  this->colorspace = JCS_GRAYSCALE ;

  return 0 ;
}

void
JPEG::setMM ( float mwidth, float mheight, float mdepth )
{
  if ( mwidth > 0.0f && mheight <= 0.0f ) {
    this->mm_width = mwidth ;
    this->mm_height = ( mwidth / this->width ) * this->height ;
  } else if ( mwidth <= 0.0f && mheight > 0.0f ) {
    this->mm_height = mheight ;
    this->mm_width = ( mheight / this->height ) * this->width ;
  } else if ( mwidth > 0.0f && mheight > 0.0f ) {
    this->mm_width = mwidth ;
    this->mm_height = mheight ;
  } else {
    this->mm_width = this->width ;
    this->mm_height = this->height ;
  }

  if ( mdepth > -65535.0f ) {
    this->mm_depth = mdepth ;
  } else {
    this->mm_depth = 1.0f ;
  }

  cout << "setMM: mm_width = " << this->mm_width <<
  " mm_height = " << this->mm_height <<
  " mm_depth = " << this->mm_depth << endl ;
}

int
JPEG::exportToSTL ( const char * filename, int skip_x, int skip_y,
		    bool binary )
{
  ofstream out_file ;
  out_file.open ( filename ) ;

  if ( ! out_file.is_open() ) {
    cerr << filename << " nie zosta otwarty!" << endl ;
    return 1 ;
  }

  unsigned int rows = this->height / skip_y ;
  unsigned int cols = this->width / skip_x ;
  float * row1 = new float [ cols ] ;
  float * row2 = new float [ cols ] ;
  float * jmp = NULL ;
  float wpc = this->mm_width / cols ; // width per column
  float hpr = this->mm_height / rows ; // height per row
  float dps ; // depth per shade
  float mh2 = this->mm_height / 2.0f ;
  float mw2 = 0.0f - this->mm_width / 2.0f ;
  unsigned char mingray = 255, maxgray = 0 ;
  //  Vertex va, vb, vc ;
  Facet facet ;

  for ( int i = 0 ; i < this->width ; ++ i )
    for ( int j = 0 ; j < this->height ; ++ j )
      {
	unsigned char g = this->buffer[ ( j * this->width ) + i ] ;
	if ( mingray > g ) mingray = g ;
	if ( maxgray < g ) maxgray = g ;
      }

  dps = this->mm_depth / ( maxgray - mingray ) ; 

  /* set fixed */
  //  out_file << fixed << setprecision ( 3 ) ;
  //  out_file << fixed ; //<< setprecision ( 3 ) ;

  cout << "exportToSTL: skip_x = " << skip_x <<
    ", skip_y = " << skip_y <<
    ", dps = " << this->mm_depth << "/" << maxgray - mingray << ", " << dps <<
    " Mesh " << cols - 1 << " x " << rows - 1 <<
    ", Triangles = " << ( cols - 1 ) * ( rows - 1 ) * 2 << endl ;

  if( ! binary ) {
    out_file << "solid " << filename << endl ;
  } else {
    string fn = filename ;
    int t = ( cols - 1 ) * ( rows - 1 ) * 2 ;
    //int * pt = & t ;

    //for ( int i = 0 ; i < 80 ; ++ i ) fn[ i ] = ( char ) 0 ;
    //filename.copy( fn, 80 ) ;
    fn.resize( 80, ' ' ) ;

    out_file.write( fn.c_str(), 80 ) ;
    out_file.write( ( char * ) & t, 4 ) ;
  }

  float mm_prev_row = mh2, mm_prev_col = mw2 ;

  for ( unsigned int col = 0 ; col < cols ; ++ col ) {
    row1[ col ] = ( this->buffer[ skip_x * col ] - mingray ) * dps ;
  }

  for ( unsigned int row = 1 ; row < rows ; ++ row ) {
    for ( unsigned int col = 0 ; col < cols ; ++ col ) {
      row2[ col ] = ( this->buffer[ ( skip_y * row * this->width ) +
				    skip_x * col ] - mingray ) * dps;
    }

    float mm_row = mh2 - row * hpr ;
    mm_prev_col = mw2 ;

    for ( unsigned int col = 1 ; col < cols ; ++ col ) {
      float mm_col = mw2 + col * wpc ;

      facet.vertexes[ 0 ].set( mm_col, mm_row, row2[ col ] ) ;
      facet.vertexes[ 1 ].set( mm_col, mm_prev_row, row1[ col ] ) ;
      facet.vertexes[ 2 ].set( mm_prev_col, mm_prev_row, row1[ col - 1 ] ) ;
      facet.normalize() ;

      if( ! binary ) {
	out_file << "facet normal " << facet.normal.x <<
	  " " << facet.normal.y << " " << facet.normal.z << endl ;
	out_file << " outer loop" << endl ;

	out_file << "  vertex " << facet.vertexes[ 0 ].x <<
	  " " << facet.vertexes[ 0 ].y <<
	  " " << facet.vertexes[ 0 ].z << endl ;
	out_file << "  vertex " << facet.vertexes[ 1 ].x <<
	  " " << facet.vertexes[ 1 ].y <<
	  " " << facet.vertexes[ 1 ].z << endl ;
	out_file << "  vertex " << facet.vertexes[ 2 ].x <<
	  " " << facet.vertexes[ 2 ].y <<
	  " " << facet.vertexes[ 2 ].z << endl ;

	out_file << " endloop" << endl ;
	out_file << "endfacet" << endl ;
      } else {
	out_file.write( ( char * ) & facet.normal.x , 4 ) ;
	out_file.write( ( char * ) & facet.normal.y , 4 ) ;
	out_file.write( ( char * ) & facet.normal.z , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 0 ].x , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 0 ].y , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 0 ].z , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 1 ].x , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 1 ].y , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 1 ].z , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 2 ].x , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 2 ].y , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 2 ].z , 4 ) ;
	out_file.write( ( char * ) "  " , 2 ) ;
      }

      facet.vertexes[ 0 ].set( mm_prev_col, mm_row, row2[ col - 1 ] ) ;
      facet.vertexes[ 1 ].set( mm_col, mm_row, row2[ col ] ) ;
      facet.vertexes[ 2 ].set( mm_prev_col, mm_prev_row, row1[ col - 1 ] ) ;
      facet.normalize() ;

      if( ! binary ) {
	out_file << "facet normal " << facet.normal.x <<
	  " " << facet.normal.y << " " << facet.normal.z << endl ;
	out_file << " outer loop" << endl ;

	out_file << "  vertex " << facet.vertexes[ 0 ].x <<
	  " " << facet.vertexes[ 0 ].y <<
	  " " << facet.vertexes[ 0 ].z << endl ;
	out_file << "  vertex " << facet.vertexes[ 1 ].x <<
	  " " << facet.vertexes[ 1 ].y <<
	  " " << facet.vertexes[ 1 ].z << endl ;
	out_file << "  vertex " << facet.vertexes[ 2 ].x <<
	  " " << facet.vertexes[ 2 ].y <<
	  " " << facet.vertexes[ 2 ].z << endl ;

	out_file << " endloop" << endl ;
	out_file << "endfacet" << endl ;
      } else {	// if binary
	out_file.write( ( char * ) & facet.normal.x , 4 ) ;
	out_file.write( ( char * ) & facet.normal.y , 4 ) ;
	out_file.write( ( char * ) & facet.normal.z , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 0 ].x , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 0 ].y , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 0 ].z , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 1 ].x , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 1 ].y , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 1 ].z , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 2 ].x , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 2 ].y , 4 ) ;
	out_file.write( ( char * ) & facet.vertexes[ 2 ].z , 4 ) ;
	out_file.write( ( char * ) "  " , 2 ) ;
      }		// if ! binary

      mm_prev_col = mm_col ;
    }
    jmp = row1 ;
    row1 = row2 ;
    row2 = jmp ;

    mm_prev_row = mm_row ;
  }

  if( ! binary ) {
    out_file << "endsolid " << filename << endl ;
  }

  delete [] row1 ;
  delete [] row2 ;

  out_file.close() ;

  return 0 ;
}

int
JPEG::save ( const char * filename )
{
  struct jpeg_compress_struct cinfo ;
  struct jpeg_error_mgr jerr ;

  FILE * outfile ;		/* target file */
  JSAMPROW row_pointer[ 1 ] ;	/* pointer to JSAMPLE row[s] */
  int row_stride ;		/* physical row width in image buffer */


  cinfo.err = jpeg_std_error( & jerr ) ;
  jpeg_create_compress( & cinfo ) ;

  if ( ( outfile = fopen( filename , "w+" ) ) == NULL ) {
    cerr << "Nie mog otworzy " << filename << endl ;
    exit( 1 ) ;
  }
  jpeg_stdio_dest( & cinfo , outfile ) ;

  cinfo.image_width = this->width ; 	/* image width and height, in pixels */
  cinfo.image_height = this->height ;

  /* # of color components per pixel */
  cinfo.input_components = this->components ;

  /* colorspace of input image */
  cinfo.in_color_space = this->colorspace ;
  jpeg_set_defaults( & cinfo ) ;

  /* compression in percents */
  jpeg_set_quality( & cinfo , this->compress , TRUE ) ;
  jpeg_start_compress( & cinfo , TRUE ) ;

  row_stride = this->width ;	/* JSAMPLEs per row in image_buffer */

  while ( cinfo.next_scanline < cinfo.image_height ) {
    /* jpeg_write_scanlines expects an array of pointers to scanlines.
     * Here the array is only one element long, but you could pass
     * more than one scanline at a time if that's more convenient.
     */
    row_pointer[ 0 ] = & this->buffer[ cinfo.next_scanline * row_stride ] ;
    ( void ) jpeg_write_scanlines( & cinfo , row_pointer , 1 ) ;
  }

  jpeg_finish_compress( & cinfo ) ;
  fclose( outfile ) ;

  jpeg_destroy_compress( & cinfo ) ;

  return 0 ;
}

