#include <cstdlib>
#include <iostream>

#include "stl.hpp"
#include "v3d.hpp"

using namespace std;

int
STL::load ( const char * filename )
{
  float min_x, max_x, min_y, max_y, min_z, max_z ;
  max_x = max_y = max_z = -65536.0f ;
  min_x = min_y = min_z = 65536.0f ;
  ifstream in_file ;
  Facet * triangle = NULL ;
  int vertex = 0 ;
  int unknown_text = 0 ;

  in_file.open ( filename ) ;

  if ( ! in_file.is_open() ) {
    cerr << filename << " nie zosta otwarty!" << endl ;
    return 1 ;
  }

  if ( ! this->binmode ) {

    cout << "Load STL - tekstowy." << endl ;

    while ( ! in_file.eof() ) {
      string line, value ;
    
      getline ( in_file, line ) ;

      //if ( line.length() < 5 ) continue ;

      if( ! this->popFrontValue( line, value ) ) continue ;

      if ( value.compare( "facet" ) == 0 ) {
	// nowy trjkt/ zeruj dyzurny trjkt
	triangle = new Facet ;
	vertex = 0 ;
	//cout << value << " nr. " << this->triangles.size() << endl ;
      } else if ( value.compare( "outer" ) == 0 ) {
	/// ???
	//cout << value << endl ;
      } else if ( value.compare( "vertex" ) == 0 ) {
	// wczytaj trzy vertex-y
	// aktualizuj BoundingBox
	string sx, sy, sz ;
	float fx = 0.0f, fy = 0.0f, fz = 0.0f ;

	this->popFrontValue( line, sx ) ;
	this->popFrontValue( line, sy ) ;
	this->popFrontValue( line, sz ) ;

	stringstream( sx ) >> fx ;
	stringstream( sy ) >> fy ;
	stringstream( sz ) >> fz ;

	if( min_x > fx ) min_x = fx ;
	if( min_y > fy ) min_y = fy ;
	if( min_z > fz ) min_z = fz ;
	if( max_x < fx ) max_x = fx ;
	if( max_y < fy ) max_y = fy ;
	if( max_z < fz ) max_z = fz ;

	if ( vertex < 3 )
	  triangle->vertexes[ vertex ].set( fx, fy, fz ) ;
	else
	  cerr << "Zbyt wiele vertex-w " << this->triangles.size() <<
	    " trjkta." << endl ;

	++ vertex ;

	//cout << value << " x=" << fx << " y=" << fy << " z=" << fz << endl ;
      } else if ( value.compare( "endloop" ) == 0 ) {
	// Wstaw odczytane wartoci vertex-w do trjkta
	//cout << value << endl ;
      } else if ( value.compare( "endfacet" ) == 0 ) {
	// trjkt (wypeniony) wrzuci do listy trjktw
	this->triangles.push_back( triangle ) ;
	//cout << value << " nr. " << this->triangles.size() << endl ;
      } else if ( value.compare( "endsolid" ) == 0 ) {
	// Oblicz wymiary rzeczywiste
	this->mm_width = max_x - min_x ;
	this->mm_height = max_y - min_y ;
	this->mm_depth = max_z - min_z ;
	this->min.x = min_x ;
	this->min.y = min_y ;
	this->min.z = min_z ;

      } else if ( value.compare( "solid" ) == 0 ) {
	// dla porzdku
      } else {
	// Co niespodziewanego, czyli zapewne binarny STL
	++ unknown_text ;
	//if ( unknown_text > 2 )
      }	// endif value.compare
    }	// while in_file
  } else {	// Binarny STL
    //int nr = 0 ;
    char name [ 81 ] ;
    //unsigned char b4 [ 4 ] ;
    int facet_number = 0 ;
    int * f_n = & facet_number ;

    cout << "Load STL - binarny." << endl ;
    /*
    cout << "Short int: " << sizeof( short int ) <<
      ", Int: " << sizeof( int ) << ", Long int" << sizeof( long int ) <<
      ", Float: " << sizeof( float ) << ", Double: " << sizeof( double ) <<
      endl ;
    */

    in_file.get( name, 81 ) ;
    in_file.read( ( char * ) f_n, 4 ) ;

    cout << name << " Facet number: " << facet_number << endl ;

    for ( int ln = 0 ; ln < facet_number ; ++ ln ) {
      unsigned char trash [ 100 ] ;
      float fx = 0.0f, fy = 0.0f, fz = 0.0f ;
      float * pfx = & fx, * pfy = & fy, * pfz = & fz ;

      triangle = new Facet ;

      in_file.read( ( char * ) trash, 12 ) ; // pomijam normaln

      for ( int vertex = 0 ; vertex < 3 ; ++ vertex ) {
	in_file.read( ( char * ) pfx, 4 ) ;
	in_file.read( ( char * ) pfy, 4 ) ;
	in_file.read( ( char * ) pfz, 4 ) ;

	if( min_x > fx ) min_x = fx ;
	if( min_y > fy ) min_y = fy ;
	if( min_z > fz ) min_z = fz ;
	if( max_x < fx ) max_x = fx ;
	if( max_y < fy ) max_y = fy ;
	if( max_z < fz ) max_z = fz ;

	triangle->vertexes[ vertex ].set( fx, fy, fz ) ;

	//cout << " V(" << fx << ", " << fy << ", " << fz << ") " ;

      }	// for vertex
      //cout << endl ;

      in_file.read( ( char * ) trash, 2 ) ; // pomijam "fill bytes"

      this->triangles.push_back( triangle ) ;

    }	// for ln

    this->mm_width = max_x - min_x ;
    this->mm_height = max_y - min_y ;
    this->mm_depth = max_z - min_z ;
    this->min.x = min_x ;
    this->min.y = min_y ;
    this->min.z = min_z ;

  }	// if ! this->binmode

  in_file.close() ;

  cout << "Load STL " << filename << " BBox " <<
    this->mm_width << " " << this->mm_height << " " << this->mm_depth <<
    " triangles " << this->triangles.size() << endl ;

  return unknown_text ;
}

int
STL::save ( const char * filename )
{
  cout << "Save STL. Niedziaajca opcja." << endl ;

  return 0 ;
}

void
STL::setPX ( float fwidth, float fheight, float fdepth )
{
#define MINSIZE 400

  int pwidth = ( int ) roundf( fwidth ) ;
  int pheight = ( int ) roundf( fheight ) ;
  int pdepth = ( int ) roundf( fdepth ) ;

  float mwidth = this->mm_width, mheight = this->mm_height ;
  
  if( mwidth <= mheight ) {
    if( mwidth < MINSIZE ) {
      float mul = MINSIZE / mwidth ;
      mwidth *= mul ;
      mheight *= mul ;
    }
    } else {
      if( mheight < MINSIZE ) {
	float mul = MINSIZE / mheight ;
	mwidth *= mul ;
	mheight *= mul ;
      }
    }

  
  if ( pwidth > 0 && pheight <= 0 ) {

    if ( pwidth < MINSIZE ) {
      this->width = ( int ) ceilf( mwidth ) ;
      this->height = ( int ) ceilf( mheight ) ;
      cout << "setPX: case 1a." << endl ;
    } else {
      this->width = pwidth ;
      this->height = ( int ) ceilf( ( pwidth / this->mm_width ) *
				    this->mm_height ) ;
      cout << "setPX: case 1b." << endl ;
    }
  } else if ( pwidth <= 0 && pheight > 0 ) {

    if ( pheight < MINSIZE ) {
      this->width = ( int ) ceilf( mwidth ) ;
      this->height = ( int ) ceilf( mheight ) ;
      cout << "setPX: case 2a." << endl ;
    } else {
      this->height = pheight ;
      this->width = ( int ) ceilf ( ( pheight / this->mm_height ) *
				    this->mm_width ) ;
      cout << "setPX: case 2b." << endl ;
    }
  } else if ( pwidth > 0 && pheight > 0 ) {

    if( pwidth >= MINSIZE && pheight >= MINSIZE ) {

      this->width = pwidth ;
      this->height = pheight ;
      cout << "setPX: case 3a." << endl ;

    } else if( pwidth < MINSIZE && pheight >= MINSIZE ) {

      // pwidth na MINSIZE, pheight przeskalowa
      float ph = ( float ) pheight / ( float ) pwidth ;

      this->width = MINSIZE ;
      this->height = ( int ) ceilf( MINSIZE * ph ) ;
      cout << "setPX: case 3b." << endl ;

    } else if( pwidth >= MINSIZE && pheight < MINSIZE ) {

      // pheight na MINSIZE, pwidth przeskalowa
      float pw = ( float ) pwidth / ( float ) pheight ;

      this->height = MINSIZE ;
      this->width = ( int ) ceilf( MINSIZE * pw ) ;
      cout << "setPX: case 3c." << endl ;

    } else {
      if ( pwidth <= pheight ) {

	float ph = ( float ) pheight / ( float ) pwidth ;

	this->width = MINSIZE ;
	this->height = ( int ) ceilf( MINSIZE * ph ) ;
	cout << "setPX: case 3d1." << endl ;

      } else {

	float pw = ( float ) pwidth / ( float ) pheight ;

	this->height = MINSIZE ;
	this->width = ( int ) ceilf( MINSIZE * pw ) ;
	cout << "setPX: case 3d2." << endl ;

      }
    }

  } else { // ustaw krtszy bok na MINSIZE
    
    this->width = ( int ) ceilf( mwidth ) ;
    this->height = ( int ) ceilf( mheight ) ;
    cout << "setPX: case 4." << endl ;
    
  }

  // Poprawne dla omiobitowej skali szaroci ( dla 16b max = 65535 )  
  if ( pdepth > 0 ) {
    if ( pdepth < 2 )
      this->depth = 2 ;
    else if ( pdepth > 255 )
      this->depth = 255 ;
    else
      this->depth = pdepth ;
  } else {
    this->depth = 255 ;
  }

  cerr << "setPX: width = " << this->width <<
    " height = " << this->height <<
    " depth = " << this->depth << endl ;

#undef MINSIZE
}

bool
STL::popFrontValue ( string & str, string & val )
{
  size_t posv = str.find_first_not_of( " \t", 0 ) ;

  if ( posv != val.npos ) str.erase( 0, posv ) ;

  if ( str.length() > 0 ) {
    posv = str.find_first_of( " \t\n", 0 ) ;

    if( posv != val.npos ) {
      val = str.substr( 0, posv ) ;
      str.erase( 0, posv ) ;
    } else
      val = str ;
  } else
    return false ;
  
  return true ;
}

#include "jpg.hpp"

int
STL::exportToJPEG ( const char * filename, int cmpr )
{
  cout << "ExportToJPEG " << filename << endl ;

  //   przygotuj bufor. Dla JPEG TYLKO 1bpp
  this->buffer = new unsigned char [ this->width * this->height ] ;

  for( int i = 0 ; i < this->width * this->height ; ++ i )
    this->buffer[ i ] = 0 ;

  float mmpp_x, mmpp_y, mmpp_z ;

  mmpp_x = ( this->width - 1 ) / this->mm_width ;
  mmpp_y = ( this->height - 1 ) / this->mm_height ;
  mmpp_z = this->depth / this->mm_depth ;

  // 0. Dla kadego trjkta
  for ( vecFacet::iterator it = this->triangles.begin() ;
	it != this->triangles.end() ; ++ it ) {
  // 1. Posortuj vertex-y rosnco wedug wsprzdnej y

    Vertex A = ( * it )->vertexes[ 0 ] ;
    Vertex B = ( * it )->vertexes[ 1 ] ;
    Vertex C = ( * it )->vertexes[ 2 ] ;
    Vertex vz, S, E ;
    if( A.y > B.y ) { vz = A ; A = B ; B = vz ; }
    if( B.y > C.y ) { vz = B ; B = C ; C = vz ; }
    if( A.y > B.y ) { vz = A ; A = B ; B = vz ; }

  // 2. Oblicz wsprzdne pikselowe A.x B.x C.x A.y B.y C.y
  //    i szaroci A.z B.z C.z

    A.x = ( ( A.x - this->min.x ) * mmpp_x ) ;
    B.x = ( ( B.x - this->min.x ) * mmpp_x ) ;
    C.x = ( ( C.x - this->min.x ) * mmpp_x ) ;
    A.y = ( ( A.y - this->min.y ) * mmpp_y ) ;
    B.y = ( ( B.y - this->min.y ) * mmpp_y ) ;
    C.y = ( ( C.y - this->min.y ) * mmpp_y ) ;
    A.z = ( ( A.z - this->min.z ) * mmpp_z ) ;
    B.z = ( ( B.z - this->min.z ) * mmpp_z ) ;
    C.z = ( ( C.z - this->min.z ) * mmpp_z ) ;
    /*
    cout << "(" << A.x << ", " << A.y << ", " << A.z <<
      ") (" << B.x << ", " << B.y << ", " << B.z <<
      ") (" << C.x << ", " << C.y << ", " << C.z << ")" << endl;
    */
    float dx1, dx2, dx3, dz1, dz2, dz3 ;

    // 3. Oblicz wspczynniki

    if ( B.y - A.y > 0 ) {
      dx1 = ( B.x - A.x ) / ( B.y - A.y ) ;
      dz1 = ( B.z - A.z ) / ( B.y - A.y ) ;
    } else {
      dx1 = B.x - A.x ;
      dz1 = B.z - A.z ;
    }
    if ( C.y - A.y > 0 ) {
      dx2 = ( C.x - A.x ) / ( C.y - A.y ) ;
      dz2 = ( C.z - A.z ) / ( C.y - A.y ) ;
    } else {
      dx2 = 0 ;
      dz2 = 0 ;
    }
    if ( C.y - B.y > 0 ) {
      dx3 = ( C.x - B.x ) / ( C.y - B.y ) ;
      dz3 = ( C.z - B.z ) / ( C.y - B.y ) ;
    } else {
      dx3 = 0 ;
      dz3 = 0 ;
    }

    // 4. Wyrysuj wszystko co si da

    S = E = A ;

    if ( dx1 > dx2 ) {

	for ( ; S.y < B.y ; S.y += 1.0f, E.y += 1.0f, S.x += dx2, E.x += dx1,
		S.z += dz2, E.z += dz1 )
 	  this->drawHLineJPEG( S.x, E.x, irndf( S.y ), S.z, E.z ) ;

	E = B ;

	for ( ; S.y < C.y ; S.y += 1.0f, E.y += 1.0f, S.x += dx2, E.x += dx3,
		S.z += dz2, E.z += dz3 )
	  this->drawHLineJPEG( S.x, E.x, irndf( S.y ), S.z, E.z ) ;

    } else {

	for ( ; S.y < B.y ; S.y += 1.0f, E.y += 1.0f, S.x += dx1, E.x += dx2,
		S.z += dz1, E.z += dz2 )
	  this->drawHLineJPEG( S.x, E.x, irndf( S.y ), S.z, E.z ) ;

	S = B ;

	for ( ; S.y < C.y ; S.y += 1.0f, E.y += 1.0f, S.x += dx3, E.x += dx2,
		S.z += dz3, E.z += dz2 )
	  this->drawHLineJPEG( S.x, E.x, irndf( S.y ), S.z, E.z ) ;

    }
  }	// for it

  // Stwrz obiekt JPEG, skopiuj mu bufor, warto bpp, itd
  // ( Zorganizowa to bez JPEG-a!!! )
  JPEG * outjpeg = new JPEG( this->width, this->height, 1,
			     JCS_GRAYSCALE, cmpr ) ;
  outjpeg->copyBuffer( this->buffer, 0, this->width * this->height ) ;
  // wywoaj jego metod save.
  outjpeg->save( filename ) ;

  delete outjpeg ;

  return 0 ;
}

inline void
STL::drawHLineJPEG( float x1, float x2, int y, float z1, float z2 )
{
  if ( x2 - x1 == 0 ) return ;

  if ( x2 < x1 ) {
    float b = x2 ; x2 = x1 ; x1 = b ;
    b = z2 ; z2 = z1 ; z1 = b ;
  }

  // bez tego obraz bdzie odwrcony
  y = this->height - y ;
 
  int pos = this->width * y ;
  float dz = ( z2 - z1 ) / ( x2 - x1 ) ;

  //  cout << "drawHLineJPEG: " << z << ", " << dz << endl ;

  for ( int i = pos + irndf( x1 ) ; i <= pos + irndf( x2 ) ; ++ i ) {

    if ( irndf( z1 ) > 255 ) {
      cerr << "z! " ;
      z1 = 255.0f ;
    } else if ( irndf( z1 ) < 0 ) {
      cerr << "-z! " ;
      z1 = 0.0f ;
    }

    unsigned char tb = irndf( z1 ) ;

    if ( tb > this->buffer[ i ] )
      this->buffer[ i ] = tb ;

    z1 += dz ;
  }
}

inline int irndf( float a )
{
  if( a >= 0.0f )
    return ( int ) ( a + 0.5f ) ;
  else
    return ( int ) ( a - 0.5f ) ;

  return ( int ) a ;
}
