#include <cstdlib>
#include <iostream>
#include <sstream>
#include <iomanip>

#include <getopt.h>

#include "jpg.hpp"
#include "stl.hpp"

using namespace std;

int
main ( int argc, char ** argv )
{

  int skip_x = 4, skip_y = 4, cmpr = -1 ;
  float mm_width = -1.0f, mm_height = -1.0f, mm_depth = -65536.0f ;
  bool save = false, binary = false ;
  const char * const short_options = "x:y:w:h:d:c:sib" ;
  const struct option long_options[] = {
    { "xskip", 1, NULL, 'x' },
    { "yskip", 1, NULL, 'y' },
    { "width", 1, NULL, 'w' },
    { "height", 1, NULL, 'h' },
    { "depth", 1, NULL, 'd' },
    { "compres", 1, NULL, 'c' },
    { "save", 0, NULL, 's' },
    { "help", 0, NULL, 'i' },
    { "binary", 0, NULL, 'b' },
    /*{ "", , NULL, '' },*/
    { NULL, 0, NULL, 0 },
  } ;
  int next_option ;

  do {
    next_option = getopt_long( argc, argv,
                               short_options, long_options, NULL ) ;

    switch ( next_option ) {
    case 'x' : /* x skip - JPG */
      stringstream( optarg ) >> skip_x ;
      break ;

    case 'y' : /* y skip - JPG */
      stringstream( optarg ) >> skip_y ;
      break ;

    case 'w' : /* width of result - JPG */
      stringstream( optarg ) >> mm_width ;
      break ;

    case 'h' : /* height of result - JPG */
      stringstream( optarg ) >> mm_height ;
      break ;

    case 'd' : /* depht of result - JPG */
      stringstream( optarg ) >> mm_depth ;
      break ;

    case 'c' : /* compression - BOTH */
      stringstream( optarg ) >> cmpr ;
      break ;

    case 's' : /* save the grayscale - JPG */
      save = true ;
      break ;

    case 'b' : /* binary STL - JPG */
      binary = true ;
      break ;

    case 'i' : /* Print info */
      printHelp() ;
      return 0 ;
      break ;

    case '?' : /* unknown option */
      cout << "\nNieznana opcja\nPomoc: ras2stl -i\n" <<
	endl ;
      return 1 ;
      break ;

    case -1 : /* end of options */
      break ;

    default :
      cerr << "getopt returned character code " << next_option << endl ;
      break ;

    }
  } while ( next_option != -1 ) ;

  FILE * tmp ;
  string fname, outfname ;
  /*
    if ( optind < argc ) {
    cout << "non-option ARGV-elements: " ;
    while (optind < argc)
    cout << argv [ optind ++ ] ;
    cout << endl ;
    }
  */

  // Pierwsza nieprzetworzona opcja to pewnie nazwa pliku
  if ( optind < argc ) {
    fname = argv [ optind ] ;
  }

  if ( (tmp = fopen ( fname.c_str() , "r" )) == NULL ) {
    cerr << "Nie mog otworzy pliku wejciowego " << fname.c_str() << endl ;
    return 1 ;
  }

  fclose ( tmp ) ;

  // Rozpoznanie rozszerzenia

  outfname = fname ;
  size_t dot = outfname.rfind( '.', outfname.length() ) ;
  string ext ;
  if ( dot != outfname.npos ) {
    ext = outfname.substr( dot + 1, outfname.length() - dot ) ;
  } else {
    cerr << "Nieznany typ pliku!" << endl ;
    return 1 ;
  }

  // Do boju

  if ( ext.compare( "jpg" ) == 0 || ext.compare( "JPG" ) == 0 ||
       ext.compare( "jpeg" ) == 0 || ext.compare( "JPEG" ) == 0 ) {

    outfname.replace( dot, outfname.length() - dot, ".stl", 0, 4 ) ;

    cout << "Konwersja JPEG na STL. Plik " << outfname << endl ;

    JPEG * jpg = new JPEG ;

    jpg->load ( fname.c_str() ) ;
    if( cmpr > 0 ) jpg->setCompression ( cmpr ) ;
    jpg->RGBtoGrayscale () ;

    jpg->setMM ( mm_width, mm_height, mm_depth ) ;
    jpg->exportToSTL ( outfname.c_str(), skip_x, skip_y, binary ) ;

    if ( save ) jpg->save ( fname.c_str() ) ;

    delete jpg ;

  } else if ( ext.compare( "stl" ) == 0 || ext.compare( "STL" ) == 0 ) {

    outfname.replace( dot, outfname.length() - dot, ".jpg", 0, 4 ) ;

    cout << "Konwersja STL na JPEG. Plik " << outfname << endl ;

    STL * stl = new STL ;
    
    stl->load ( fname.c_str() ) ;

    if ( stl->numOfTriangles() < 1 ) {
      stl->toBinmode() ;
      stl->load ( fname.c_str() ) ;
    }
    
    if ( stl->numOfTriangles() > 0 ) {

      stl->setPX ( mm_width, mm_height, mm_depth ) ;
      if ( cmpr < 1 ) cmpr = 95 ;
      stl->exportToJPEG ( outfname.c_str(), cmpr ) ;

      // if binary stl->toBinmode else stl->toTextmode
      if ( save ) stl->save ( fname.c_str() ) ;

    } else
      cout << "Nic nie wczytano z " << fname << endl ;

    delete stl ;

  }

  return 0 ;

}

void
printHelp ( void )
{
  /*
    { "xskip", 1, NULL, 'x' },
    { "yskip", 1, NULL, 'y' },
    { "width", 1, NULL, 'w' },
    { "height", 1, NULL, 'h' },
    { "depth", 1, NULL, 'd' },
    { "compres", 1, NULL, 'c' },
    { "save", 0, NULL, 's' },
    { "help", 0, NULL, 'i' },
  */

  cout << "triangularez [OPCJE] nazwa_pliku.\n" << endl ;


  cout << "Opcje oglne:" << endl ;
  cout << "  -i --help		Wypisuje ten tekst." << endl ;


  cout << "\nOpcje konwersji JPEG -> STL:\n" << endl ;
  cout << "  -x --xskip	PX	Piksele JPEG-a do pominicia w poziomie.\n" <<
    "			Ta opcja zmniejszy rozmiar pliku STL." << endl ;
  cout << "  -y --yskip	PX	Piksele JPEG-a do pominicia w pionie.\n" <<
    "\t\t\tTa opcja zmniejszy rozmiar pliku STL." << endl ;
  cout << "  -w --width	MM	Szeroko obiektu zapisanego" <<
    " w STL w milimetrach." << endl ;
  cout << "  -h --height	MM	Wysoko obiektu zapisanego" <<
    " w STL w milimetrach.\n" <<
    "\t\t\tUWAGA: Jeli ma by zachowana proporcja szerokoci\n" <<
    "\t\t\tdo wysokoi naley poda tylko -h lub tylko -w." <<
    endl ;
  cout << "  -d --depth	MM	Gboko obiektu zapisanego"
    " w STL w milimetrach." << endl ;
  cout << "  -s --save		Przekonwertuj JPEG rdowy na" <<
    " odcienie szaroci\n\t\t\ti zapisz pod star nazw." << endl ;
  cout << "  -c --compres	PROC	Stopie kompresji JPEG-a" <<
    " w procentach.\n\t\t\tUWAGA: Dziaa z opcj -s." << endl ;
  cout << "  -b --binary	\tGeneruj binarny STL." << endl ;

  cout << "\nOpcje konwersji STL -> JPEG:\n" << endl ;
  cout << "  -w --width	PX	Szeroko JPEG-a w pikselach." << endl ;
  cout << "  -h --height	PX	Wysoko JPEG-a w pikselach." << endl ;
  cout << "  -d --depth	GRAY	Ilo odcieni (od 2 do 255)" << endl ;

  cout << endl ;
}
